@@ -72,6 +72,9 @@ example, if our first expectation had been 3 instead of 4, we'd see:
 
 The main functions you'll need are:
 
+- `trace` works like `print`, but shows some extra information, and you
+  can use it as part of a larger expression. Use this for figuring out
+  what's going wrong when your expectations aren't met.
 - `testCase` establishes a test case, and subsequent expectations are
   attached to the most recent test case.
 - `expectResult`, `expectOutputContains`, and/or `expectCustom` to
@@ -100,11 +103,14 @@ The main functions you'll need are:
 Note: currently, if test causes an error, you just get the error, and
 the testing machinery doesn't do anything about that.
 TODO: Some way to do something about that?
+
+TODO: Workaround for tracing in interactive console?
+TODO: Prevent crash if expr_src is unavailable when tracing...
 """
 
 # TODO: Cache compiled ASTs!
 
-__version__ = "0.9.3"
+__version__ = "0.11"
 
 import sys
 import inspect
@@ -114,18 +120,6 @@ import io
 import types
 import builtins
 
-# Monkey-patch get_source_segment into earlier Pythons in a bad way
-if not hasattr(ast, "get_source_segment"):
-    ast.get_source_segment = lambda src, node: (
-        src.splitlines()
-            [node.lineno - 1]
-            [node.col_offset:]
-    )
-    print(
-        "Warning: ast module does not have a get_source_segment"
-        " function; expression values will include extra code. Use"
-        " Python version 3.8 or greater for ideal results."
-    )
 
 #---------#
 # Globals #
@@ -289,7 +283,12 @@ def captureOutput():
     """
     global OUTPUT_CAPTURE, ORIGINAL_INPUT_FN
     if OUTPUT_CAPTURE is None or sys.stdout != OUTPUT_CAPTURE:
-        ORIGINAL_INPUT_FN = builtins.input
+        # Note: we don't want to set up a recursive input function!
+        # Note: this can happen despite the if above when someone else
+        # re-captures output in between redundant attempts to capture
+        # output
+        if builtins.input != _echoing_input:
+            ORIGINAL_INPUT_FN = builtins.input
         builtins.input = _echoing_input
         OUTPUT_CAPTURE = CapturingStream()
         OUTPUT_CAPTURE.install()
@@ -533,8 +532,8 @@ def testCase(expr):
         source code (something you should probably avoid anyway), none of
         the calls should execute more times than another when that line
         is executed (it's difficult to violate this, but examples include
-        the use of `testCase` multiple times on line line within
-        generator or if/else expressions)
+        the use of `testCase` multiple times on one line within generator
+        or if/else expressions)
     2. None of the following components of the expression passed to
         `testCase` should have side effects when evaluated:
         - Attribute accesses
@@ -691,7 +690,7 @@ def create_failure_message(
 def expr_details(context):
     """
     Returns a pair of strings containing base and extra details for an
-    expectation. Both strings will start with a newline.
+    expectation.
     """
     # Expression that was evaluated
     expr = context.get("expr_src", "???")
@@ -701,44 +700,67 @@ def expr_details(context):
     extra_msg = ""
 
     # Base message
+    msg += f"In expression {short_expr}, values were:"
+    vdict = context.get("values", {})
     if context.get("relevant"):
-        msg += f"In expression {short_expr}, values were:"
-        for key in sorted(
+        show = sorted(
             context["relevant"],
-            key=lambda fragment: expr.index(fragment)
-        ):
-            val = context.get("values", {}).get(key, "???")
-            msg += f"\n  {key} = {val}"
+            key=lambda fragment: (expr.index(fragment), len(fragment))
+        )
     else:
-        msg += f"In expression {short_expr}, values were:"
-        vdict = context.get("values", {})
-        for key in sorted(
+        show = sorted(
             vdict.keys(),
-            key=lambda fragment: expr.index(fragment)
-        ):
-            val = context.get("values", {}).get(key, "???")
-            msg += f"\n  {key} = {val}"
+            key=lambda fragment: (expr.index(fragment), len(fragment))
+        )
+
+    longs = []
+    for key in show:
+        if key in vdict:
+            val = repr(vdict[key])
+        else:
+            val = "???"
+
+        entry = f"  {key} = {val}"
+        fits = ellipsis(entry)
+        msg += '\n' + fits
+        if fits != entry:
+            longs.append(entry)
 
     # Extra message
     if short_expr != expr:
-        if not extra_msg.endswith('\n'):
+        if extra_msg != "" and not extra_msg.endswith('\n'):
             extra_msg += '\n'
         extra_msg += f"Full expression:\n  {expr}"
     extra_values = sorted(
         [
             key
-            for key in context.get("values", {}).keys()
+            for key in vdict.keys()
             if key not in context.get("relevant", [])
         ],
-        key=lambda fragment: expr.index(fragment)
+        key=lambda fragment: (expr.index(fragment), len(fragment))
     )
     if context.get("relevant") and extra_values:
-        if not extra_msg.endswith('\n'):
+        if extra_msg != "" and not extra_msg.endswith('\n'):
             extra_msg += '\n'
         extra_msg += "Extra values:"
         for ev in extra_values:
-            val = context.get("values", {}).get(ev, "???")
-            extra_msg += f"\n  {ev} = {val}"
+            if ev in vdict:
+                val = repr(vdict[ev])
+            else:
+                val = "???"
+
+            entry = f"  {ev} = {val}"
+            fits = ellipsis(entry)
+            extra_msg += '\n' + fits
+            if fits != entry:
+                longs.append(entry)
+
+    if longs:
+        if extra_msg != "" and not extra_msg.endswith('\n'):
+            extra_msg += '\n'
+        extra_msg += "Full values:"
+        for entry in longs:
+            extra_msg += '\n' + entry
 
     return msg, extra_msg
 
@@ -785,21 +807,24 @@ def check_expectation(exp, set_success=True):
             return True
         else:
             # Construct base report
-            msg = (
+            base_msg = (
                 f"Result {short_result} was different from the"
                 f" expected value {short_expected}."
             )
 
             # Construct extra-detailed message
-            extra_msg = (
-                f"Full result:\n  {full_result}\n"
-                f"Full expectation:\n  {full_expected}"
-            )
+            extra_msg = ""
+            if short_result != full_result:
+                extra_msg += f"Full result:\n  {full_result}"
+            if short_expected != full_expected:
+                if extra_msg != "":
+                    extra_msg += '\n'
+                extra_msg += f"Full expectation:\n  {full_expected}"
 
             msg = create_failure_message(
                 exp["tag"],
                 context,
-                msg,
+                base_msg,
                 extra_msg
             )
             print_message(msg, color='1;31' if COLORS else None)
@@ -1150,6 +1175,333 @@ def trace(expr):
 # Reverse evaluation machinery #
 #------------------------------#
 
+def get_src_index(src, lineno, col_offset):
+    """
+    Turns a line number and column offset into an absolute index into
+    the given source string, assuming length-1 newlines.
+    """
+    lines = src.splitlines()
+    above = lines[:lineno - 1]
+    return sum(len(line) for line in above) + len(above) + col_offset
+
+
+def test_gsr():
+    """Tests for get_src_index."""
+    s = 'a\nb\nc'
+    assert get_src_index(s, 1, 0) == 0
+    assert get_src_index(s, 2, 0) == 2
+    assert get_src_index(s, 3, 0) == 4
+    assert s[get_src_index(s, 1, 0)] == 'a'
+    assert s[get_src_index(s, 2, 0)] == 'b'
+    assert s[get_src_index(s, 3, 0)] == 'c'
+
+
+def find_identifier_end(code, start_index):
+    """
+    Given a code string and an index in that string which is the start
+    of an identifier, returns the index of the end of that identifier.
+    """
+    at = start_index + 1
+    while at < len(code):
+        ch = code[at]
+        if not ch.isalpha() and not ch.isdigit() and ch != '_':
+            break
+        at += 1
+    return at - 1
+
+
+def test_find_identifier_end():
+    """Tests for find_identifier_end."""
+    assert find_identifier_end("abc.xyz", 0) == 2
+    assert find_identifier_end("abc.xyz", 1) == 2
+    assert find_identifier_end("abc.xyz", 2) == 2
+    assert find_identifier_end("abc.xyz", 4) == 6
+    assert find_identifier_end("abc.xyz", 5) == 6
+    assert find_identifier_end("abc.xyz", 6) == 6
+    assert find_identifier_end("abc_xyz123", 0) == 9
+    assert find_identifier_end("abc xyz123", 0) == 2
+    assert find_identifier_end("abc xyz123", 4) == 9
+    assert find_identifier_end("x", 0) == 0
+    assert find_identifier_end("  x", 2) == 2
+    assert find_identifier_end("  xyz1", 2) == 5
+    s = "def abc(def):\n  print(xyz)\n"
+    assert find_identifier_end(s, 0) == 2
+    assert find_identifier_end(s, 4) == 6
+    assert find_identifier_end(s, 8) == 10
+    assert find_identifier_end(s, 16) == 20
+    assert find_identifier_end(s, 22) == 24
+
+
+def unquoted_enumerate(src, start_index):
+    """
+    A generator that yields index, character pairs from the given code
+    string, skipping quotation marks and the strings that they delimit,
+    including triple-quotes and respecting backslash-escapes within
+    strings.
+    """
+    quote = None
+    at = start_index
+
+    while at < len(src):
+        char = src[at]
+
+        # skip escaped characters in quoted strings
+        if quote and char == '\\':
+            # (thank goodness I don't have to worry about r-strings)
+            at += 2
+            continue
+
+        # handle quoted strings
+        elif char == '"' or char == "'":
+            if quote == char:
+                quote = None # single end quote
+                at += 1
+                continue
+            elif src[at:at + 3] in ('"""', "'''"):
+                tq = src[at:at + 3]
+                at += 3 # going to skip these no matter what
+                if tq == quote or tq[0] == quote:
+                    # Ending triple-quote, or matching triple-quote at
+                    # end of single-quoted string = ending quote +
+                    # empty string
+                    quote = None
+                    continue
+                else:
+                    if quote:
+                        # triple quote of other kind inside single or
+                        # triple quoted string
+                        continue
+                    else:
+                        quote = tq
+                        continue
+            elif quote is None:
+                # opening single quote
+                quote = char
+                at += 1
+                continue
+            else:
+                # single quote inside other quotes
+                at += 1
+                continue
+
+        # Non-quote characters in quoted strings
+        elif quote:
+            at += 1
+            continue
+
+        else:
+            yield (at, char)
+            at += 1
+            continue
+
+
+def test_unquoted_enumerate():
+    """Tests for unquoted_enumerate."""
+    uqe = unquoted_enumerate
+    assert list(uqe("abc'123'", 0)) == list(zip(range(3), "abc"))
+    assert list(uqe("'abc'123", 0)) == list(zip(range(5, 8), "123"))
+    assert list(uqe("'abc'123''", 0)) == list(zip(range(5, 8), "123"))
+    assert list(uqe("'abc'123''", 1)) == [(1, 'a'), (2, 'b'), (3, 'c')]
+    mls = "'''\na\nb\nc'''\ndef"
+    assert list(uqe(mls, 0)) == list(zip(range(12, 16), "\ndef"))
+    tqs = '"""\'\'\'ab\'\'\'\'""" cd'
+    assert list(uqe(tqs, 0)) == [(15, ' '), (16, 'c'), (17, 'd')]
+    rqs = "a'b'''c\"\"\"'''\"d\"''''\"\"\"e'''\"\"\"f\"\"\"'''"
+    print(f"X: '{rqs[23]}'", file=sys.stderr)
+    assert list(uqe(rqs, 0)) == [(0, 'a'), (6, 'c'), (23, 'e')]
+    assert list(uqe(rqs, 6)) == [(6, 'c'), (23, 'e')]
+    bss = "a'\\'b\\''c"
+    assert list(uqe(bss, 0)) == [(0, 'a'), (8, 'c')]
+    mqs = "'\"a'b\""
+    assert list(uqe(mqs, 0)) == [(4, 'b')]
+
+
+def find_nth_attribute_period(code, start_index, n):
+    """
+    Given a string of Python code and a start index within that string,
+    finds the nth period character (counting from first = zero) after
+    that start point, but only considers periods which are used for
+    attribute access, i.e., periods outside of quoted strings and which
+    are not part of ellipses. Returns the index within the string of the
+    period that it found. A period at the start index (if there is one)
+    will be counted. Returns None if there are not enough periods in the
+    code. If the start index is inside a quoted string, things will get
+    weird, and the results will probably be wrong.
+    """
+    for (at, char) in unquoted_enumerate(code, start_index):
+        if char == '.':
+            if code[at - 1:at] == '.' or code[at + 1:at + 2] == '.':
+                # part of an ellipsis, so ignore it
+                continue
+            else:
+                n -= 1
+                if n < 0:
+                    break
+
+    # Did we hit the end of the string before counting below 0?
+    if n < 0:
+        return at
+    else:
+        return None
+
+
+def test_find_nth_attribute_period():
+    """Tests for find_nth_attribute_period."""
+    assert find_nth_attribute_period("a.b", 0, 0) == 1
+    assert find_nth_attribute_period("a.b", 0, 1) is None
+    assert find_nth_attribute_period("a.b", 0, 100) is None
+    assert find_nth_attribute_period("a.b.c", 0, 1) == 3
+    assert find_nth_attribute_period("a.b.cde.f", 0, 1) == 3
+    assert find_nth_attribute_period("a.b.cde.f", 0, 2) == 7
+    s = "a.b, c.d, 'e.f', g.h"
+    assert find_nth_attribute_period(s, 0, 0) == 1
+    assert find_nth_attribute_period(s, 0, 1) == 6
+    assert find_nth_attribute_period(s, 0, 2) == 18
+    assert find_nth_attribute_period(s, 0, 3) is None
+    assert find_nth_attribute_period(s, 0, 3) is None
+    assert find_nth_attribute_period(s, 1, 0) == 1
+    assert find_nth_attribute_period(s, 2, 0) == 6
+    assert find_nth_attribute_period(s, 6, 0) == 6
+    assert find_nth_attribute_period(s, 7, 0) == 18
+    assert find_nth_attribute_period(s, 15, 0) == 18
+
+
+def find_closing_item(code, start_index, openclose='()'):
+    """
+    Given a string of Python code, a starting index where there's an
+    open paren, bracket, etc., and a 2-character string containing the
+    opening and closing delimiters of interest (parentheses by default),
+    returns the index of the matching closing delimiter, or None if the
+    opening delimiter is unclosed. Note that the given code must not
+    contain syntax errors, or the behavior will be undefined.
+
+    Does NOT work with quotations marks (single or double).
+    """
+    level = 1
+    open_delim = openclose[0]
+    close_delim = openclose[1]
+    for at, char in unquoted_enumerate(code, start_index + 1):
+        # Non-quoted open delimiters
+        if char == open_delim:
+            level += 1
+
+        # Non-quoted close delimiters
+        elif char == close_delim:
+            level -= 1
+            if level < 1:
+                break
+
+        # Everything else: ignore it
+
+    if level == 0:
+        return at
+    else:
+        return None
+
+
+def test_find_closing_item():
+    """Tests for find_closing_item."""
+    assert find_closing_item('()', 0, '()') == 1
+    assert find_closing_item('()', 0) == 1
+    assert find_closing_item('(())', 0, '()') == 3
+    assert find_closing_item('(())', 1, '()') == 2
+    assert find_closing_item('((word))', 0, '()') == 7
+    assert find_closing_item('((word))', 1, '()') == 6
+    assert find_closing_item('(("(("))', 0, '()') == 7
+    assert find_closing_item('(("(("))', 1, '()') == 6
+    assert find_closing_item('(("))"))', 0, '()') == 7
+    assert find_closing_item('(("))"))', 1, '()') == 6
+    assert find_closing_item('(()())', 0, '()') == 5
+    assert find_closing_item('(()())', 1, '()') == 2
+    assert find_closing_item('(()())', 3, '()') == 4
+    assert find_closing_item('(""")(\n""")', 0, '()') == 10
+    assert find_closing_item("\"abc(\" + ('''def''')", 9, '()') == 19
+    assert find_closing_item("\"abc(\" + ('''def''')", 0, '()') is None
+    assert find_closing_item("\"abc(\" + ('''def''')", 4, '()') is None
+    assert find_closing_item("(()", 0, '()') is None
+    assert find_closing_item("(()", 1, '()') == 2
+    assert find_closing_item("()(", 0, '()') == 1
+    assert find_closing_item("()(", 2, '()') is None
+    assert find_closing_item("[]", 0, '[]') == 1
+    assert find_closing_item("[]", 0) is None
+    assert find_closing_item("{}", 0, '{}') == 1
+    assert find_closing_item("aabb", 0, 'ab') == 3
+
+
+def get_expr_src(src, call_node):
+    """
+    Gets the string containing the source code for the expression passed
+    to a function call, given the string source of the file that defines
+    the function and the AST node for the function call.
+    """
+    # Find the child node for the first (and only) argument
+    arg_expr = call_node.args[0]
+
+    # If get_source_segment is available, use that
+    if hasattr(ast, "get_source_segment"):
+        return ast.get_source_segment(src, arg_expr)
+    else:
+        # We're going to have to do this ourself: find the start of the
+        # expression and state-machine to find a matching paren
+        start = get_src_index(src, call_node.lineno, call_node.col_offset)
+        open_paren = src.index('(', start)
+        end = find_closing_item(src, open_paren, '()')
+        # Note: can't be None because that would have been a SyntaxError
+        return src[open_paren + 1:end]
+
+
+def get_ref_src(src, node):
+    """
+    Gets the string containing the source code for a variable reference,
+    attribute, or subscript.
+    """
+    # Use get_source_segment if it's available
+    if hasattr(ast, "get_source_segment"):
+        return ast.get_source_segment(src, node)
+    else:
+        # We're going to have to do this ourself: find the start of the
+        # expression and state-machine to find its end
+        start = get_src_index(src, node.lineno, node.col_offset)
+
+        # Figure out the end point
+        if isinstance(node, ast.Attribute):
+            # Find sub-attributes so we can count syntactic periods to
+            # figure out where the name part begins to get the span
+            inner_period_count = 0
+            for node in ast.walk(node):
+                if isinstance(node, ast.Attribute):
+                    inner_period_count += 1
+            inner_period_count -= 1 # for the node itself
+            dot = find_nth_attribute_period(src, start, inner_period_count)
+            end = find_identifier_end(src, dot + 1)
+
+        elif isinstance(node, ast.Name):
+            # It's just an identifier so we can find the end
+            end = find_identifier_end(src, start)
+
+        elif isinstance(node, ast.Subscript):
+            # Find start of sub-expression so we can find opening brace
+            # and then match it to find the end
+            inner = node.slice
+            if isinstance(inner, ast.Slice):
+                pass
+            elif hasattr(ast, "Index") and isinstance(inner, ast.Index):
+                # 3.7 Index has a "value"
+                inner = inner.value
+            elif hasattr(ast, "ExtSlice") and isinstance(inner, ast.ExtSlice):
+                # 3.7 ExtSlice has "dims"
+                inner = inner.dims[0]
+            else:
+                raise TypeError(
+                    f"Unexpected subscript slice type {type(inner)} for"
+                    f" node:\n{ast.dump(node)}"
+                )
+            sub_start = get_src_index(src, inner.lineno, inner.col_offset)
+            end = find_closing_item(src, sub_start - 1, "[]")
+
+        return src[start:end + 1]
+
+
 def deepish_copy(obj, memo=None):
     """
     Returns the deepest possible copy of the given object, using
@@ -1211,8 +1563,8 @@ def deepish_copy(obj, memo=None):
 def get_external_calling_frame():
     """
     Uses the inspect module to get a reference to the stack frame which
-    called into the `expectations` module. Returns None if it can't find
-    an appropriate call frame in the current stack.
+    called into the `optimism` module. Returns None if it can't find an
+    appropriate call frame in the current stack.
 
     Remember to del the result after you're done with it, so that
     garbage doesn't pile up.
@@ -1301,7 +1653,10 @@ def walk_ast_in_order(node):
         # Note: the node itself will be yielded LAST
         if isinstance(node, (ast.Module, ast.Interactive, ast.Expression)):
             yield from walk_ast_in_order(node.body)
-        elif isinstance(node, ast.FunctionType):
+        elif (
+            hasattr(ast, "FunctionType")
+        and isinstance(node, ast.FunctionType)
+        ):
             yield from walk_ast_in_order(node.argtypes)
             yield from walk_ast_in_order(node.returns)
         elif isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
@@ -1359,12 +1714,13 @@ def walk_ast_in_order(node):
 
         elif isinstance(node, ast.BoolOp):
             yield from walk_ast_in_order(node.values)
-        elif isinstance(node, ast.NamedExpr):
+        elif hasattr(ast, "NamedExpr") and isinstance(node, ast.NamedExpr):
             yield from walk_ast_in_order(node.value)
             yield from walk_ast_in_order(node.target)
-        elif isinstance(node, ast.BinOp):
-            yield from walk_ast_in_order(node.left)
-            yield from walk_ast_in_order(node.right)
+        # TODO: HERE
+        #elif isinstance(node, ast.BinOp):
+        #    yield from walk_ast_in_order(node.left)
+        #    yield from walk_ast_in_order(node.right)
         elif isinstance(node, ast.UnaryOp):
             yield from walk_ast_in_order(node.operand)
         elif isinstance(node, ast.Lambda):
@@ -1418,7 +1774,8 @@ def walk_ast_in_order(node):
         elif isinstance(node, ast.arguments):
             yield from walk_ast_in_order(node.defaults)
             yield from walk_ast_in_order(node.kw_defaults)
-            yield from walk_ast_in_order(node.posonlyargs)
+            if hasattr(node, "posonlyargs"):
+                yield from walk_ast_in_order(node.posonlyargs)
             yield from walk_ast_in_order(node.args)
             yield from walk_ast_in_order(node.vararg)
             yield from walk_ast_in_order(node.kwonlyargs)
@@ -1431,6 +1788,10 @@ def walk_ast_in_order(node):
             yield from walk_ast_in_order(node.context_expr)
             yield from walk_ast_in_order(node.optional_vars)
         # alias and typeignore have no executable members
+        # TODO: HERE
+        elif isinstance(node, ast.BinOp):
+            yield from walk_ast_in_order(node.left)
+            yield from walk_ast_in_order(node.right)
 
         # Finally, yield this node itself
         yield node
@@ -1628,7 +1989,7 @@ def get_my_context(function_or_name):
         assign_parents(arg_expr)
 
         # Source code for the expression
-        expr_src = ast.get_source_segment(src, arg_expr)
+        expr_src = get_expr_src(src, match)
 
         # Prepare our result dictionary
         result = {
@@ -1648,7 +2009,7 @@ def get_my_context(function_or_name):
                 node,
                 (ast.Attribute, ast.Subscript, ast.Name)
             ):
-                key = ast.get_source_segment(src, node)
+                key = get_ref_src(src, node)
                 if key not in result["values"]:
                     # Don't re-evaluate multiply-reference expressions
                     # Note: we assume they won't take on multiple
